// AD9850 DDS chip driver
// OH2NLT Juha Niinikoski 02.04.2004
// AD9851 test code added 26.08.2004
// Modified for JUMA-TRX2 & Microchip C30 24.11.2660
/*
 The Tuning Word = ( Frequency << 33 ) / ( Reference Oscillator * 6 )
 O/P frequency = Tuning Word * ( Reference Oscillator * 6 ) / 2 ^ 32
 Note that the output frequency is twice the signal frequency, there is a cross-coupled divide by 2 stage
 that is used to generate the I and Q phases.
 
 Example: Crystal Oscillator Frequency = 30MHz
 For an output frequency for 10MHz receive:
 Tuning Word	= 10,000,000 * 2 ^ 33 / 30,000,000 * 6
				= 10,000,000 * 8,589,934,592 / 180,000,000
				= 477,218,583 (Decimal) or 0x1C71C717

 Output Frequency	= 477,218,583 * 180,000,000 / 4,294,967,296
					= 19,999,999.772Hz

 Suppose the 30MHz oscillator is 1kHz high.
 If the displayed frequency is 10MHz, then the actual carrier frequency is:
				= 477,218,583 * 180,006,000 / 4,294,967,296
				= 20,000,666.439Hz

 In order to correct this, the reference oscillator frequency has to be corrected, thus the tuning word is:
				= 10,000,000 * 2 ^ 33 / 30,001,000 * 6
				= 10,000,000 * 8,589,934,592 / 180,006,000
				= 477,202,681 (Decimal) or 0x1C7188F9

 Output frequency	= 477,202,681 * 180,006,000 / 4,294,967,296
					= 19,999,999.971Hz

 Modified the operation of dds_load(). Although the multiplication by the constant 8589934592ULL may seem
 counter-intuitive, as the processor contains a hardware multiplier and divider, this operation actually
 only takes a single machine operation, and is faster than the equivalent << 33. Also altered the structure
 DDS_Register so that it now contains a long long. A.Ryan - 5B4AIY - 23/AUG/2015
*/

#define AD9851_CFG		0x01	// Control & phase bits = 0x01, 6x multiplier enabled

#include "juma-trx2.h"			// JUMA-TRX2 hardware specific definitions
#include "trx2_eeprom.h"		// JUMA-TRX2 Data EEPROM structures

// Calibration Data
extern union
	{
	struct calval calval;
	unsigned int ee[sizeof(struct calval) / 2];
	} cal;
/*
 This declares a union called: DDS, consisting of a single object, the structure
 DDS_Registers, defined in juma_trx2.h, which is referred to by the name: Registers,
 but can also be referred to as an array of 8 unsigned characters called: AD9850[].
 */
union										// AD9851 frequency data. AD9851[0] = LSB
	{
	struct DDS_Registers Registers;
	unsigned char AD9851[8];
	} DDS;

// AD9851 data transfer
void dds_send_one(unsigned char data)		// Send one byte LSB first
	{
	int bit_counter = 8;					// Changed to type int from type char - 5B4AIY

	do 	{
		DDS_SER = (data & 0x01);			// This is surely simpler - 5B4AIY
		data >>= 1;							// Move next bit into place
		DDS_W_CLK = 1;						// Generate clock pulse, at a clock rate of 7.5MHz this corresponds
		DDS_W_CLK = 0;						// to a clock pulse width of 133nS, the minimum is 3.5nS - 5B4AIY
		} while (--bit_counter);
	}

void dds_load(unsigned long freq)			// Load AD9850 from buffer
	{
	static unsigned long old_freq;
	int i = 0;

	if(freq != old_freq)					// If frequency has changed,
		{									// then calculate the new frequency word, and load it.
		old_freq = freq;
		DDS.Registers.Frequency = ((long long)freq * 8589934592ULL) / (long long)cal.calval.ref_osc;

		do	{								// Send 40 bits to synthesiser,
			dds_send_one(DDS.AD9851[i]);	// 32 frequency data bits, (4 bytes),
			} while (++i < 4);

		dds_send_one(AD9851_CFG);			// and 1 configuration byte.
		DDS_FQ_UD = 1;						// Generate update pulse
		DDS_FQ_UD = 0;
		}
	}

/*
 Initialise the AD9851 interface
 When the AD9851 initialises after being reset, it comes up in the default parallel load mode. In order
 to change this to the serial load mode used in this application it is necessary to strobe a control word
 into the W0 register immediately after the reset. This control word must have Bit-0 set to force the use
 of the 6x clock multiplier, and Bit-1 set to force the use of the serial mode. This is achieved by hard
 wiring the data bit pins to +5V, and performing a frequency update. Immediately afterwards a serial frequency
 command should be entered, even if it is all zeros, to flush the 40-bit frequency register buffer to ensure
 that there is valid data in the buffer.
*/
void dds_init(void)
	{
	DDS_RESET = 1;							// Reset AD9850
	DDS_SER = 0;
	DDS_RESET = 0;
	DDS_FQ_UD = 1;							// Generate Frequency Update pulse,
	DDS_FQ_UD = 0;							// and set the serial mode.
	dds_load(DEFAULT_FREQ_A);				// Initial load to establish working point.
	}

