/*
 JUMA-TRX2 driver for ISD1700 Voice recorder chip
 Software driver for SPI2 interface
 Juha Niinikoski, OH2NLT 10.03.2007
 MPLAB C30 version
 MCU dsPIC30F6014A

 Memory banks changed, voice memory option installed flag 17.03.2007
 UART2 added 02.01.2008

 Added include file stdio.h, and cast to pointer to char to eliminate warning errors. A.Ryan - 5B4AIY - 9/AUG/11
 Modified tx_rx_16_spi2() and tx_rx_56_spi2() so that they no longer return any data. In all previous versions the return data was discarded anyway.
 Modified tx_rx_32_spi2() so that the returned data is masked to 8 bits. The return data is only ever used during the device initialisation to
 determine the device ID when called by init_isd1700() A.Ryan - 5B4AIY - 23/AUG/2013
 Added array slot_time[] to retrieve the values for the maximum slot times. A.Ryan - 5B4AIY - 25/SEP/2013
*/

#include <stdio.h>					// Added this include file to eliminate errors - 5B4AIY
#include "juma-trx2.h"				// JUMA-TRX2 hardware specific definitions
#include "trx2_eeprom.h"			// Get EEPROM structure definitions

// ISD17240 Voice Recorder I/O is defined in juma-trx2.h
// ISD1700 commands
// 2-byte commands, second byte = 0x00
#define PU				0x01		// Power Up
#define STOP			0x02		// Stop current operation
#define RESET			0x03		// Reset the device
#define CLR_INT			0x04		// Clear int & EOM bit
#define PD				0x07		// Power Down
#define PLAY			0x40		// Play from current location
#define REC				0x41		// Record from current location
#define ERASE			0x42		// Erase current message to EOM
#define G_ERASE			0x43		// Erase all messages
#define WR_NVCFG		0x46		// Write APC to NVCFG
#define LD_NVCFG		0x47		// Load APC from NVCFG
#define FWD				0x48		// Forward playback pointer to next message
#define CHK_MEM			0x49		// Check circular memory
#define EXTCLK			0x4A		// Enable/disable ext clock mode

// 3-byte commands
#define RD_STATUS		0x05		// Returns current row counter & operational status
#define DEVID			0x09		// Read device ID register
#define WR_APC1			0x45		// Write D11:D0 into the APC reg, VOL from pin
#define WR_APC2			0x65		// Write D11:D0 into the APC reg, VOL from APC D2:D0

// 4-byte commands
#define RD_PLAY_PTR		0x06		// Returns status bits & current row counter & playback pointer
#define RD_REC_PTR		0x08		// Returns status bits & current row counter & record pointer
#define RD_APC			0x44		// Returns status bits & current row counter & APC register

// 7-byte commands
#define SET_PLAY		0x80		// Play start S10:S0 to end E10:E0 or EOM or STOP
#define SET_REC			0x81		// Record start S10:S0 to end E10:E0 or STOP
#define SET_ERASE		0x82		// Erase from start S10:S0 to end E10:E0

// Device ID Mask
#define ID_MASK			0xF8		// The device ID code occupies bits 7 through 3 of the ID byte.

// Command modifier bits
#define	LED				0x10		// ISD1700 LED indicator on

// SR0 bits
#define	SR0_CMD_ERR		(1 << 0)	// Command error
#define	SR0_FULL		(1 << 1)	// Memory full
#define	SR0_PU			(1 << 2)	// Device is powered and operating in SPI mode
#define	SR0_EOM			(1 << 3)	// This bit is detected when EOM is detected, cleared by CLR_INT command
#define	SR0_INT			(1 << 4)	// This bit is set when current operation is done, cleared by CLR_INT command 

// SR1 bits
#define	SR1_RDY			(1 << 0)	// Device is ready to accept new command
#define	SR1_ERASE		(1 << 1)	// Current operation is erase
#define	SR1_PLAY		(1 << 2)	// Current operation is playback
#define	SR1_REC			(1 << 3)	// Current operation is recording
#define	SR1_SE1			(1 << 4)	// SE1 is recorded
#define	SR1_SE2			(1 << 5)	// SE2 is recorded
#define	SR1_SE3			(1 << 6)	// SE3 is recorded
#define	SR1_SE4			(1 << 7)	// SE4 is recorded

// APC register bits
#define APC_VOL0		(1 << 0)	// Volume control, 000 = max
#define APC_VOL1		(1 << 1)	// -4dB / bit
#define APC_VOL2		(1 << 2)
#define APC_MONITOR		(1 << 3)	// Enable input signal to output during record
#define APC_MIX_INPUT	(1 << 4)	// Input Mix control
#define APC_SE_EDITING	(1 << 5)	// Disable Sound Effect editing
#define APC_SPI_FT		(1 << 6)	// nFT bit in SPI mode, 0 = feed through
#define APC_AUX			(1 << 7)	// AUD/AUX output select, 1 = AUX
// Note D11:D8 shifted bit#-8
#define APC_PWM_SPK		(1 << 0)	// D8, PWM speaker amp disable, 1 = disable
#define APC_PU_AOUT		(1 << 1)	// D9, Analog output disable in PU mode, 1 = disable
#define APC_VALERT		(1 << 2)	// D10, Voice alert on/off, 1 = off
#define APC_EOM_ENA		(1 << 3)	// D11, EOM enable, 1 = EOM enabled

// Some APC register bit combination definitions
// APC low data must be shifted << 8 and ORed with WR_APC2 command

// APC high bits D11:D8 are same for all commands
// PWM_SPK = 1(disabled), PU_AOUT = 0(enabled), VALERT = 1(off), EOM_ENA = 1
#define APC_HI_init	(APC_PWM_SPK | APC_VALERT | APC_EOM_ENA)

// Initial state / Record from RX / message TX on the air
// Vol = max, Monitor = 1(enabled), Mix = 0, SE_Edit = 0, SPI_FT = 0(on), AUX = 1
#define APC_LO_init	(WR_APC2 | ((APC_MONITOR | APC_AUX) << 8))

// Record from Microphone
// SPI_FT = 1(off) = Mic REC, MONITOR = 1(disabled) 
#define APC_LO_mic (WR_APC2 | ((APC_SPI_FT | APC_AUX) << 8))

// Replay message (Play)
// SPI_FT = 1(off), MONITOR = 1(disabled)
#define APC_LO_play	(WR_APC2 | ((APC_MONITOR | APC_SPI_FT | APC_AUX) << 8))

// External References
// External Functions
// System Calibration Data
extern struct
	{
	struct calval calval;
	} cal;

// User Configuration Data
extern struct
	{
	struct defval defval;
	} eeprom;

extern int kbhit(void);				// UART1 input check
extern void ClearUSART1queue(void);	// Flush UART1 buffer
extern int kbhit2(void);			// UART2 input check
extern void ClearUSART2queue(void);	// Flush UART2 buffer

extern void display_line(int, const char *);
extern void lcd_cmd(unsigned char);
extern void lcd_putst(register const char *);
extern void draw_s_meter(int);
extern int get_pwr(void);
extern void beep(int, int);
extern void bl_on(void);
extern long int_round(long, long);
extern int convert_adc12(unsigned int);

// Alarms
extern void check_alarm(void);

// Timing
extern void ms_delay(unsigned int);

// External Variables
extern unsigned char lcdpbuff[];	// LCD print buffer
extern int ms_counter;
extern int isd_1700_installed;		// Voice memory board found
extern int sq_timer;				// Squelch timer
extern int alarm_flag;
extern int pwr_meter;
extern int s_meter;

// ISD17240 memory division table for JUMA-TRX2 application
// Resolution is 119.166mS/bit
const int ISD_MEM[10][2] = {
							{0x010, 0x3FF},	// [0] = MEM 0, Len = 0x3EF, 1007, 120s
							{0x400, 0x4AF},	// [1] = MEM 1, Len = 0x0AF, 175, 20s
							{0x4B0, 0x507},	// [2] = MEM 2, Len = 0x057, 87, 10s
							{0x508, 0x55F},	// [3] = MEM 3, Len = 0x057, 87, 10s
							{0x560, 0x5B7},	// [4] = MEM 4, Len = 0x057, 87, 10s
							{0x5B8, 0x60F},	// [5] = MEM 5, Len = 0x057, 87, 10s
							{0x610, 0x667},	// [6] = MEM 6, Len = 0x057, 87, 10s
							{0x668, 0x6BF},	// [7] = MEM 7, Len = 0x057, 87, 10s
							{0x6C0, 0x717},	// [8] = MEM 8, Len = 0x057, 87, 10s
							{0x718, 0x78F}	// [9] = MEM 9, Len = 0x078, 119, 14s
							};

const unsigned char slot_time[] = {
							120,			// 0
							20,				// 1
							10,				// 2
							10,				// 3
							10,				// 4
							10,				// 5
							10,				// 6
							10,				// 7
							10,				// 8
							14				// 9
							};

// ISD1700 driver variables
int isd_start_addr;					// Voice memory start address
int isd_end_addr;					// Voice memory end address
int VM_Mode = IDLE;					// Voice Memory Mode -1 = Record, +1 = Play, 0 = Idle
int seconds = 0;					// Free running 1-sec timer

// ISD1700 Standard Prompts
const char Elapsed_Time[] = {"Time:%11i"};
const char Play_Memory[] = {"PLAY Memory :%3i"};
const char TX_Memory[] = {"TX M:%1i"};
const char REC_Mic_Memory[] = {"REC Mic Mem :%3i"};
const char REC_RX_Memory[] = {"RX M:%1i Erasing.."};

// Helper functions
// Exchange SPI2 data primitive
// The return value is only used when called from tx_rx_32_spi2()
unsigned int exchg_data_spi2(unsigned int data)
	{
	int bit_counter = 16;
	unsigned int rx_data = 0;

// Exchange data with ISD1700
	do	{
		ISD_SCLK = 0;				// Generate negative clock pulse

// Set MOSI bit. Start from bit #0, LSB first
		if(data & 0x0001) ISD_MOSI = 1;		// Evaluate bit
		else ISD_MOSI = 0;
//		ISD_MOSI = (data & 0x0001);			// This takes 6 bytes more code! A.Ryan 24/JAN/2013
//		ISD_MOSI = (data & 0x0001) ? 1 : 0;	// This also takes 6 bytes more code! A.Ryan 24/JAN/2013

		data >>= 1;					// Adjust data for next round

// Test MISO bit & shift it into RX data shift register
		rx_data >>= 1;				// Shift old bit in

		if(ISD_MISO)				// Get bit from AFEDRI8201 when clk high
			rx_data |= 0x8000;

		ISD_SCLK = 1;				// Set clock high
		} while(--bit_counter);

	return rx_data;					// Return with RX data
	}

// Send / Receive 16 bit data from SPI 2
void tx_rx_16_spi2(unsigned int data)
	{
	ISD_SS = 0;						// Select chip
	exchg_data_spi2(data);			// Do exchange, discard returned data
	ISD_SS = 1;						// Deselect chip
	ISD_MOSI = 0;					// ISD1700 spec !
	}

// General Stop
void stop(void)
	{
	tx_rx_16_spi2(STOP | LED);		// Stop command
	VM_Mode = IDLE;					// Set mode to idle
	}

void check_stop(void)				// Check if STOP signal 
	{
	if(Serial_Mode == SERIAL_VOICE_MEMORY)	// UART1 mode = Ext KB, Voice Memory
		{
		if(kbhit())					// If there is a character(s) waiting from UART1,
			{
			ClearUSART1queue();		// Flush the buffer
			stop();
			return;
			}
		}
	else if(kbhit2())				// Else, if there is a character(s) waiting from UART2,
		{
		ClearUSART2queue();			// Flush the buffer
		stop();
		return;
		}

	check_alarm();

	if(alarm_flag)					// Whilst this could have been combined with the PWR switch this would
		{							// have involved a delay, which is undesirable for an alarm response.
		stop();
		return;
		}

	if(PWR_SW)						// If PWR button is pressed
		{
		ms_delay(BUTTON_DEBOUNCE);
		stop();
		}
	}

void _disp_meter(int dsp_flag)
	{
	int s_meter_type;

	if(dsp_flag)										// Display Graphic Bar-Meter
		{
		s_meter_type = eeprom.defval.S_Meter_Type;		// Save current S-Meter type
		eeprom.defval.S_Meter_Type = GRAPHIC;			// Force GRAPHIC S-Meter
		lcd_cmd(LINE1PLUS7);

		if(PTT_IN_IO)									// RX - Display Graphic S-Meter
			draw_s_meter((int)int_round((long)convert_adc12(S_METER), (long)cal.calval.s_mtr_cal));
		else											// TX - Display Graphic RF O/P Power Meter
			{
			get_pwr();
			draw_s_meter(pwr_meter);
			}

		eeprom.defval.S_Meter_Type = s_meter_type;		// Restore original S-Meter type
		}

	sprintf((char *)lcdpbuff, Elapsed_Time, seconds);	// Display elapsed time
	display_line(LINE2, lcdpbuff);
	check_stop();
	}

// Send command and get response, 32 bit transfer
// The return data is only used when called from init_isd1700()
unsigned int tx_rx_32_spi2(unsigned int cmnd, unsigned int data)
	{
	unsigned int resp;

	ISD_SS = 0;						// Select chip
	exchg_data_spi2(cmnd);			// Send command & Data Byte 1, discard RX data
	resp = exchg_data_spi2(data);	// Read response, TX data = DataByte 2
	ISD_SS = 1;						// Deselect chip
	ISD_MOSI = 0;					// ISD1700 spec !
	return resp;
	}

// 56-bit (7 byte) memory access commands
// cmnd = 0x0000 + cmnd bits
// data1 = start address, S10:S0
// data2 = end address, E10:E0
void tx_rx_56_spi2(unsigned int cmnd, unsigned int data1, unsigned int data2)
	{
	int bit_counter = 8;

	ISD_SS = 0;						// Select chip
	exchg_data_spi2(cmnd);			// Send Command & Data Byte1, discard returned data
	exchg_data_spi2(data1);			// Transfer Start Address, discard returned data
	exchg_data_spi2(data2);			// Transfer End Address, discard returned data
	ISD_MOSI = 0;

	do	{							// Send 8 zero bits
		ISD_SCLK = 0;				// Generate clock pulse
		__asm__ volatile ("nop");
		__asm__ volatile ("nop");	// 400ns high/low spec
		ISD_SCLK = 1;
		__asm__ volatile ("nop");
		__asm__ volatile ("nop");
		} while(--bit_counter);

	ISD_SS = 1;						// deselect chip
	ISD_MOSI = 0;					// ISD1700 spec !
	}

// Init ISD1700
void init_isd1700(void)				// Initialise voice memory system
	{
	ISD_RESET = 1;					// Remove hardware reset
// Init SPI2
// This section was initially in init_spi2(), but was moved here as in-line code. Only called once.
// -------------------------------------------------------------
	ISD_SS = 1;						// Set SS high
	ISD_SCLK = 1;					// SCLK idle state = 1
	ISD_MOSI = 0;					// ISD1700 spec !
// -------------------------------------------------------------

	tx_rx_16_spi2(PU);							// Power up device
	tx_rx_32_spi2(APC_LO_init, APC_HI_init); 	// Initialise ISD17240
	VM_Mode = IDLE;								// Set mode to idle
	isd_1700_installed = FALSE;					// VM Module Not Installed

	if((ID_MASK & tx_rx_32_spi2(DEVID, 0)) == ISD_17240)
		{
		tx_rx_32_spi2(APC_LO_init, APC_HI_init);// Initialise ISD17240 (Audio in = Receiver audio)
		isd_1700_installed = TRUE;				// VM Module Installed
		}
	}

// Set Start & End address
void set_isd_addresses(int idx)
	{
	isd_start_addr = ISD_MEM[idx][0];			// Convert index to physical addresses
	isd_end_addr = ISD_MEM[idx][1];

//	printf("\n\rStart = %X, End = %X\n\r",isd_start_addr, isd_end_addr);			// Debug printouts
//	printf("\n\rStart = %i, End = %i\n\r",isd_start_addr, isd_end_addr);			// Print start & stop address
	}

// Play selected memory block
// Mode 0 = Playback, 1 = Transmit
void play_mem(int mem_num, int mode)
	{
	bl_on();

	if(mode)
		{
		start_scaf_clock();						// Audio on, if squelch closed
		sq_timer = SQ_HOLD;						// Reload squelch open timer
		}

	sprintf((char *)lcdpbuff, (mode) ? TX_Memory : Play_Memory, mem_num);
	display_line(CLEAR, lcdpbuff);
	tx_rx_32_spi2(APC_LO_play, APC_HI_init); 	// RX mute, audio from ISD1700
	set_isd_addresses(mem_num);					// Set ISD1700 addresses
	ms_counter = MS_TIMER;						// Resynch millisecond timer
	seconds = 0;								// Reset playback timer,
	VM_Mode = REPLAY;							// and set mode.
	tx_rx_16_spi2(CLR_INT);						// Clear INT signal
	tx_rx_56_spi2((SET_PLAY | LED), isd_start_addr, isd_end_addr );	// Execute command

	while(ISD_RDY) _disp_meter(mode);			// While busy, display incrementing time...

	VM_Mode = IDLE;								// Return mode to idle
	tx_rx_32_spi2(APC_LO_init, APC_HI_init); 	// Audio in = Receiver audio
	}

// Erase & Record selected memory block
// Channel = 0, Mic, Channel = 1, Receiver
void rec_mem(int mem_num, int channel)
	{
	bl_on();

	if(channel)									// Record from Receiver audio
		{
		start_scaf_clock();						// Audio on, if squelch closed
		sq_timer = SQ_HOLD;						// Reload squelch open timer
		tx_rx_32_spi2(APC_LO_init, APC_HI_init);// Audio in = Receiver audio
		}
	else										// Record from Microphone
		tx_rx_32_spi2(APC_LO_mic, APC_HI_init); // Audio in = Microphone

	sprintf((char *)lcdpbuff, (channel) ? REC_RX_Memory : REC_Mic_Memory, mem_num);
	display_line(CLEAR, lcdpbuff);
	set_isd_addresses(mem_num);					// Set ISD1700 addresses
	tx_rx_16_spi2(CLR_INT);						// Clear INT signal
	tx_rx_56_spi2((SET_ERASE | LED), isd_start_addr, isd_end_addr);	// Erase memory range

	while(ISD_RDY);								// Wait for idle state

	ms_counter = MS_TIMER;						// Resynch millisecond timer,
	seconds = (int)slot_time[mem_num];			// and set seconds timer to slot time,
	VM_Mode = RECORD;							// and set mode.
	tx_rx_16_spi2(CLR_INT);						// Clear INT signal
	tx_rx_56_spi2((SET_REC | LED), isd_start_addr, isd_end_addr );	// Start recording

	while(ISD_RDY) _disp_meter(channel);		// While busy, display remaining memory time...

	VM_Mode = IDLE;
	tx_rx_16_spi2(CLR_INT);						// Clear INT signal
	tx_rx_32_spi2(APC_LO_init, APC_HI_init); 	// Restore Audio in to Receiver audio
	}

