// Cheap DSP dsPICradio board LCD routines
// This is polling driver with 8-bit interface

// Juha Niinikoski OH2NLT 26.08.2005

// Modified for Digital RX hardware 12.09.2006
// MPLAB C30 version, OH2GWE, 2006.10.17

// IRQ mask added to LCD I/O port operations. IRQ tone output needs this protection.
// S-meter roll over corrected, interface changed to int 02.12.2006
// LCD HEX output added 09.07.2007

// Includes
#include <stdio.h>
#include <string.h>
#include "juma-trx2.h"
#include "trx2_eeprom.h"			// EEPROM storage structure

// External Functions
extern void us_delay(unsigned int);
extern void ms_delay(unsigned int);

// External Variables
extern char lcdpbuff[];				// LCD print buffer, longer than the display for prints
extern int bl_timer;
extern int band_in_use;
extern unsigned int ms_counter;

// Local variables
char scale_dots[] = {				// 4 vertical dashes, S9 marker, 3 vertical dashes (Used by draw_s_meter())
						0x00,
						0x00,
						0x00,
						0x00,
						0x00,
						0x00,
						0x00,
						0x00
						};

// Local Data
const char init_lcd[] = {			// LCD initialisation data
						0x38,		// 8-bit, 2 rows, 5x7 matrix, display on
						0x38,
						0x38,
						0x0C,		// Cursor off, no blink
						0x01,
						0x06		// Increment, no display shift
						};

const char *scale_fmt[] = {			// Indexed by index in disp_meter()
						"%6.1fW",	// 0 - WATTS
						"SWR%4.1f",	// 1 - SWR
						"%6.2fV",	// 2 - VOLTS
						"%6.2fA",	// 3 - AMPS
						"%4.1fdBm"	// 4 - DBM
						};

const double scale_factor[] = {		// Indexed by index in disp_meter()
						1000.0,		// 0 - WATTS
						100.0,		// 1 - SWR
						1000000.0,	// 2 - VOLTS
						2000000.0,	// 3 - AMPS
						0.1			// 4 - DBM
						};

const char RIT_Fmt[] = {"R%+5.2f"};

unsigned int scroll_direction = 0;

// User Configuration Data
extern struct
	{
	struct defval defval;
	unsigned int storage[sizeof(struct defval) / 2];
	} eeprom;

void clk_lcd(unsigned char ldata)			// Set data & clock both LCD controllers
	{
	__asm__ volatile ("disi #9");			// Disable interrupts for 9 machine cycles. Protect IRQ tone output to RD1 I/O
	LCD_DATA = (LCD_DATA & ~LCD_MASK) | ((ldata << LCD_SHIFT) & LCD_MASK);	// set LCD data lines
	us_delay(1);							// Extra delay for filter board latch wires
	LCD_E = 1;
	us_delay(1);
	LCD_E = 0;
	us_delay(1); 							// Test
	}

/*
 By incrementing a counter in the DO-WHILE loop the accumulated number of busy counts can be determined.
 If the accumulated count was greater than the value saved in a temporary variable in the Serial Test
 Suite, this total was updated. The accumulated maximum count was displayed with some temporary code
 in the Serial Test Suite. During the initialisation the maximum count was 100, during normal displays
 it was 2. Reloading the custom character RAM seemed to take the most time. For the Juma PA-100D, the times
 were: 98/2. The loop uses 11 3-byte instructions, and thus the instruction execution time is:
 11 * 133.33nS = 1.47uS, the delays however account for just over 2uS in total giving a loop cycle time of
 3.47uS. In the case of the PA-100D, the instruction cycle time is 135.6nS, giving a loop instruction execution
 time of 3.5uS. A.Ryan - 5B4AIY - 17/AUG/2014

 The minimum setup delay from E low to high transition before data can be read is specified as 320nS maximum,
 the minimum delay available from the timer interrupt system is 1uS.
*/
void wait_lcd_rdy(void)						// Wait until LCD module is ready
	{
	unsigned int busy;						// LCD Busy Flag

	__asm__ volatile ("disi #4");			// Disable interrupts for 4 machine cycles. Protect IRQ tone output to RD1 I/O
	LCD_TRIS |= LCD_MASK;					// Make data port input, LCD bus direction bits = 1
	LCD_RW = 1;								// Read
	LCD_RS = 0;								// Command mode

	do {
		us_delay(1);						// On first execution, this provides the setup delay for RS. On subsequent
		LCD_E = 1;							// executions, a delay to allow E to settle from Hi to Lo.
		us_delay(1);						// Provides the minimum setup delay from E to Read Data.
		busy = LCD_BUS & LCD_BUSY_FLAG;		// Read port and extract BUSY flag bit,
		LCD_E = 0;
		} while (busy);						// and test for busy.

	LCD_RW = 0;
	LCD_RS = 1;
	__asm__ volatile ("disi #4");			// Disable interrupts for 4 machine cycles. Protect IRQ tone output to RD1 I/O
	LCD_TRIS &= (~LCD_MASK);				// Return to Write mode
	}

void lcd_cmd(unsigned char cmnd)
	{
	wait_lcd_rdy();
	LCD_RS = 0;								// Switch to command mode
	clk_lcd(cmnd);							// Set cursor command + cursor position
	LCD_RS = 1;								// Switch back to data mode
	}

void initlcd(void)							// Initialise selected LCD controller
	{
	int i = 0;

	LCD_RS = 0;								// Set command mode
	LCD_RW = 0;
	ms_delay(15);							// Start delay

	do	{
		clk_lcd(init_lcd[i]);				// Write initialisation data with delays
		ms_delay(5);
		} while(++i < 6);

	LCD_RS = 1;								// Set data mode
	set_pwm4_dac(DEFAULT_CONTRAST);			// Initialise the default contrast setting
	set_pwm3_dac(DEFAULT_BL);				// and default backlighting,
	bl_timer = OFF;							// disable the backlight timer,
	clear_lcd();							// and ensure LCD memory is clear.
	}

void lcdoutch(unsigned char lcd_char)		// Output character to selected display
	{
	wait_lcd_rdy();
	clk_lcd(lcd_char);
	}

// HD44780 character generator routines
// Modified for SOLOMON LM1125SYLU1 display
void set_ch_bits(char data, char count)		// Set chgen bits. Data, number of sets
	{
	while(count--) lcdoutch(data);
	}

void set_chgen(int type)					// Write bar symbols to HD44780 character generator RAM
	{
	lcd_cmd(0x00 | ADRSET);

	if(type == 2)				// Small scale markers
		{
// CGRAM 00 - Empty block, and part = 0
		set_ch_bits(0x00, 2);	// .....
								// .....
		set_ch_bits(0x10, 1);	// @....
		set_ch_bits(0x15, 2);	// @.@.@
								// @.@.@
		set_ch_bits(0x10, 1);	// @....
		set_ch_bits(0x00, 2);	// .....
								// .....

// CGRAM 01 (part = 1)
		set_ch_bits(0x00, 2);	// .....
								// .....
		set_ch_bits(0x18, 1);	// @@...
		set_ch_bits(0x1D, 2);	// @@@.@
								// @@@.@
		set_ch_bits(0x18, 1);	// @@...
		set_ch_bits(0x00, 2);	// .....
								// .....

// CGRAM 02 (part = 2)
		set_ch_bits(0x00, 2);	// .....
								// .....
		set_ch_bits(0x1C, 1);	// @@@..
		set_ch_bits(0x1D, 2);	// @@@.@
								// @@@.@
		set_ch_bits(0x1C, 1);	// @@@..
		set_ch_bits(0x00, 2);	// .....
								// .....

// CGRAM 03	(part = 3)
		set_ch_bits(0x00, 2);	// .....
								// .....
		set_ch_bits(0x1E, 1);	// @@@@.
		set_ch_bits(0x1F, 2);	// @@@@@
								// @@@@@
		set_ch_bits(0x1E, 1);	// @@@@.
		set_ch_bits(0x00, 2);	// .....
								// .....

// CGRAM 04 (part = 4)
		set_ch_bits(0x00, 2);	// .....
								// .....
		set_ch_bits(0x1F, 4);	// @@@@@
								// @@@@@
								// @@@@@
								// @@@@@
		set_ch_bits(0x00, 2);	// .....
								// .....

// CGRAM 05 - (part = 5) Fully-Filled Block
		set_ch_bits(0x00, 2);	// .....
								// .....
		set_ch_bits(0x1F, 4);	// @@@@@
								// @@@@@
								// @@@@@
								// @@@@@
		set_ch_bits(0x00, 2);	// .....
								// .....

// CGRAM 06 - End-Of-Scale Marker
		set_ch_bits(0x00, 2);	// .....
								// .....
		set_ch_bits(0x10, 4);	// @....
								// @....
								// @....
								// @....
		set_ch_bits(0x00, 2);	// .....
								// .....

// CGRAM 07 - End-Of-Scale Marker
		set_ch_bits(0x00, 2);	// .....
								// .....
		set_ch_bits(0x10, 4);	// @....
								// @....
								// @....
								// @....
		set_ch_bits(0x00, 2);	// .....
								// .....

		scale_dots[4] = 0x00;	// Normal 'empty' block for Small Scale
		}
	else if(type == 1)			// Large Scale Markers
		{
// CGRAM 00 - Empty block, and part = 0
		set_ch_bits(0x00, 1);	// .....
		set_ch_bits(0x10, 2);	// @....
								// @....
		set_ch_bits(0x15, 2);	// @.@.@
								// @.@.@
		set_ch_bits(0x10, 2);	// @....
								// @....
		set_ch_bits(0x00, 1);	// .....

// CGRAM 01 (part = 1)
		set_ch_bits(0x00, 1);	// .....
		set_ch_bits(0x10, 1);	// @....
		set_ch_bits(0x18, 1);	// @@...
		set_ch_bits(0x1D, 2);	// @@@.@
								// @@@.@
		set_ch_bits(0x18, 1);	// @@...
		set_ch_bits(0x10, 1);	// @....
		set_ch_bits(0x00, 1);	// .....

// CGRAM 02 (part = 2)
		set_ch_bits(0x00, 1);	// .....
		set_ch_bits(0x10, 1);	// @....
		set_ch_bits(0x1C, 1);	// @@@..
		set_ch_bits(0x1D, 2);	// @@@.@
								// @@@.@
		set_ch_bits(0x1C, 1);	// @@@..
		set_ch_bits(0x10, 1);	// @....
		set_ch_bits(0x00, 1);	// .....

// CGRAM 03	(part = 3)
		set_ch_bits(0x00, 1);	// .....
		set_ch_bits(0x10, 1);	// @....
		set_ch_bits(0x1E, 1);	// @@@@.
		set_ch_bits(0x1F, 2);	// @@@@@
								// @@@@@
		set_ch_bits(0x1E, 1);	// @@@@.
		set_ch_bits(0x10, 1);	// @....
		set_ch_bits(0x00, 1);	// .....

// CGRAM 04 (part = 4)
		set_ch_bits(0x00, 1);	// .....
		set_ch_bits(0x10, 1);	// @....
		set_ch_bits(0x1F, 4);	// @@@@@
								// @@@@@
								// @@@@@
								// @@@@@
		set_ch_bits(0x10, 1);	// @....
		set_ch_bits(0x00, 1);	// .....

// CGRAM 05 - (part = 5) Fully-Filled Block
		set_ch_bits(0x00, 1);	// .....
		set_ch_bits(0x10, 1);	// @....
		set_ch_bits(0x1F, 4);	// @@@@@
								// @@@@@
								// @@@@@
								// @@@@@
		set_ch_bits(0x10, 1);	// @....
		set_ch_bits(0x00, 1);	// .....

// CGRAM 06 - End-Of-Scale Marker
		set_ch_bits(0x00, 1);	// .....
		set_ch_bits(0x10, 6);	// @....
								// @....
								// @....
								// @....
								// @....
								// @....
		set_ch_bits(0x00, 1);	// .....

// CGRAM 07 - End-Of-Scale Marker
		set_ch_bits(0x00, 1);	// .....
		set_ch_bits(0x10, 6);	// @....
								// @....
								// @....
								// @....
								// @....
								// @....
		set_ch_bits(0x00, 1);	// .....

		scale_dots[4] = 0x00;	// Normal 'empty' block for Large Scale
		}
	else						//	Original Scale
		{
//	CGRAM 00 - Empty Block and part = 0
		set_ch_bits(0x00, 2);	// .....
								// .....
		set_ch_bits(0x10, 3);	// @....
								// @....
								// @....
		set_ch_bits(0x00, 3);	// .....
								// .....
								// .....

//	CGRAM 01 - part = 1
		set_ch_bits(0x00, 1);	// .....
		set_ch_bits(0x10, 5);	// @....
								// @....
								// @....
								// @....
								// @....
		set_ch_bits(0x00, 2);	// .....
								// .....

//	CGRAM 02 - part = 2
		set_ch_bits(0x00, 1);	// .....
		set_ch_bits(0x10, 5);	// @....
								// @....
								// @....
								// @....
								// @....
		set_ch_bits(0x00, 2);	// .....
								// .....

//	CGRAM 03 - part = 3
		set_ch_bits(0x00, 1);	// .....
		set_ch_bits(0x14, 5);	// @.@..
								// @.@..
								// @.@..
								// @.@..
								// @.@..
		set_ch_bits(0x00, 2);	// .....
								// .....

//	CGRAM 04 - part = 4
		set_ch_bits(0x00, 1);	// .....
		set_ch_bits(0x14, 5);	// @.@..
								// @.@..
								// @.@..
								// @.@..
								// @.@..
		set_ch_bits(0x00, 2);	// .....
								// .....

// CGRAM 05 - Full Block and part = 5
		set_ch_bits(0x00, 1);	// .....
		set_ch_bits(0x15, 5);	// @.@.@
								// @.@.@
								// @.@.@
								// @.@.@
								// @.@.@
		set_ch_bits(0x00, 2);	// .....
								// .....

// CGRAM 06 - S-Meter S9 Marker
		set_ch_bits(0x00, 2);	// .....
								// .....
		set_ch_bits(0x10, 1);	// @....
		set_ch_bits(0x00, 1);	// .....
		set_ch_bits(0x10, 1);	// @....
		set_ch_bits(0x00, 3);	// .....
								// .....
								// .....

// CGRAM 07 - End-Of-Scale
		set_ch_bits(0x00, 2);	// .....
								// .....
		set_ch_bits(0x10, 3);	// @....
								// @....
								// @....
		set_ch_bits(0x00, 3);	// .....
								// .....
								// .....

		scale_dots[4] = 0x06;	// Special S9 marker for Original scale
		}

	clear_lcd();				// Back to normal mode (clear LCD)
	}

// Draw 16 long S-meter, scale = 0...96 (48 visible steps)
// Draw 8 long S-meter, scale = 0...48 (24 visible steps)
void draw_s_meter(int len)
	{
	int whole, part, i = 0;
	char c;

	if(len > MAX_GRAPH) len = MAX_GRAPH;	// Cut to max MAX_GRAPH = 48 (Maximum graph value)

	whole = len / FONT_W;					// Calculate graph. FONT_W = 6 whole = fully filled blocks
	part = len % FONT_W;					// part = partially filled block, index into scale_dots[] array

	do	{
		if(i < whole) c = FULL_BLOCK;		// Fully-filled block
		else if(part)						// If there is a remainder,
				{
				c = part;					// part is index into partial block character map
				part = 0;					// Clear part so that on next round, we display empty blocks.
				}
				else c = scale_dots[i];		// If no remainder, then simply empty blocks. If the scale is set to Original
											// then the S9 block is special, and is already in the scale_dots[] array.
		lcdoutch(c);
		} while (++i < DISP_LEN);

	lcdoutch(END_OF_SCALE);
	}

// General I/O functions
// Display null terminated string
void lcd_putst(register const char *str)
	{
	while(*str) lcdoutch(*str++);			// Much more concise form. - 5B4AIY
	}

// Display a line of text at a specified cursor position.
void display_line(int command, const char *text)
	{
	lcd_cmd(command);
	lcd_putst(text);
	}

// Display a screen of text.
void display_screen(const char *line_1, const char *line_2)
	{
	display_line(CLEAR, line_1);
	display_line(LINE2, line_2);
	}

/*
 This is a complete re-design of the frequency display module. A.Ryan - 5B4AIY - 05/OCT/2014
 Optimised by removing the redundant leading zero flag. A.Ryan - 5B4AIY - 31/AUG/2016

 The frequency can be displayed in one of four modes (Assume the frequency is: 12.345678MHz):
 ORIGINAL	- Two decimal points, 10Hz resolution, and leading zeros.
 12.345.67
 For frequencies below 1MHz, there will be leading zeros:
  0.000.67
 NEW		- One decimal point, 1Hz resolution, leading zero suppression, except for Hz frequencies:
 12.345678
  2.345678
  0.345678
     0.678

 FIXED-A	- One decimal point, 1Hz resolution, leading zeros
 12.345678
  0.000678

 FIXED-B	- One decimal point, 10Hz resolution, leading zeros
 12.34567
  0.00067
 For this option, to maintain the display width, a trailing space is displayed.
 This re-design saved 348 bytes.

 Display rounding occurs if this option is set in the User Configuration.

 If the frequency is converted using: sprintf(freq_data, "%08ld", freq); then the main loop cycles
 in 14.7mSec. Using the local conversion instead, the main loop cycles in 7.12mSec.
 Main loop cycle time is now 3mS to 4mS depending upon display type. See notes in adc12.c

 The display field width is 9 characters, and, using the ORIGINAL display as a template:

 Index i	:	0 1[2]2 3 4[5]5 6
 Frequency	:	1 2 . 3 4 5 . 6 7
 
 The indices in parentheses represent the decimal point location(s) (dp1, dp2) in relation to the index.

 The number of digits to display, num, is initially set to 7 to account for the normal display with a
 frequency resolution of 10Hz, but is set to 8 for the NEW and FIXED-A displays, which display to 1Hz.

 As only the ORIGINAL display uses two decimal points, for all the other displays dp1 and dp2 are set
 to the same value, 2. The exception is the NEW display where the decimal point moves from between the
 units of MHz and hundreds of kHz digits to the units of kHz and hundreds of Hz digits for frequencies
 lower than 1MHz.

 Since the FIXED-B display only uses a single decimal point, and a frequency resolution of 10Hz, a single
 trailing space is displayed to pad out the frequency field width to 9 characters.
*/
void disp_freq(long freq)
	{
	int c, i, num = 7, dp1 = 2, dp2 = 2, lz = TRUE;
	long divisor = 10000000L;

	lcd_cmd(LINE2);

	if(eeprom.defval.Freq_Display == ORIGINAL) dp2 = 5;

	if(eeprom.defval.Freq_Display == NEW_DISP)
		{
		num = 8;
		if(freq < 1000000L) dp1 = dp2 = 5;
		}

	if(eeprom.defval.Freq_Display == FIXED_A) num = 8;

	for(i = 0; i < num; i++, divisor /= 10L)
		{
		c = (int)(freq / divisor);								// Extract next digit,
		freq -= (long)c * divisor;								// and remove current decade value.
		
		if(lz && c) lz = FALSE;									// At the first non-zero digit, reset the leading zero flag.

		if(lz && !c && (i < (dp1 - 1))) c = SPACE;				// If necessary, suppress leading zeros,
		else c += ZERO;											// otherwise convert to ASCII.

		if((i == dp1) || (i == dp2))							// Insert decimal point(s), 
			lcdoutch((!band_in_use && (ms_counter > 499)) ? SPACE : '.');	// if out-of-band, then blink decimal points.

		lcdoutch((unsigned char)c);								// Display next frequency digit,
		}

	if(eeprom.defval.Freq_Display == FIXED_B) lcdoutch(SPACE);	// and add a trailing space for the FIXED-B mode.
	}
/*
  Display RIT format - new version 5B4AIY
  The display is: R+1.00 or R-1.00 or R 0.00 (Maximum offset is 1,000Hz)
  With this implementation, adjusting the offset to exactly 0Hz will display: R 0.00
  with a +1Hz offset the display will be: R+0.00, with a -1Hz offset the display will be: R-0.00,
  thus the user can set the RIT frequency to exactly zero by adjusting for no sign.
*/
void disp_rit(int x)
	{
	sprintf(lcdpbuff, RIT_Fmt, ((double)x) / 1000.0);	// The + in the format statement forces a leading +/- symbol

	if(!x) lcdpbuff[1] = SPACE;							// If the offset is exactly zero, do not display a sign.

	lcd_putst(lcdpbuff);
}

// Display meter format - new version 5B4AIY
// Replaced SWITCH/CASE construct with a more efficient indexed array construct. A.Ryan - 5B4AIY - 12/JAN/2014
void disp_meter(double value, int index)
	{
	sprintf(lcdpbuff, scale_fmt[index], value / scale_factor[index]);
	lcd_putst(lcdpbuff);
	}

/*
 The scroll_LCD() function scrolls the LCD display buffer one character at a time,
 adds the character c to the buffer, and displays the result.
*/
void scroll_LCD(int line, char c)
	{
	int i;

	if(scroll_direction)
		{
		i = 15;

		do	{
			lcdpbuff[i] = lcdpbuff[i - 1];	// Shift display right one character
			} while(--i);

		lcdpbuff[0] = c;			// Add the new character,
		}
	else
		{
		i = 0;

		do	{
			lcdpbuff[i] = lcdpbuff[i + 1];	// Shift display left one character
			} while(++i < 15);

		lcdpbuff[15] = c;			// Add the new character,
		}

	lcdpbuff[16] = 0;				// terminate the buffer,
	display_line(line, lcdpbuff);	// and display it.
	}

