// JUMA-TRX2 external serial keyboard for Voice Memory control
// Juha Niinikoski, OH2NLT 12.03.2007

// User interface "Type 2" added 14.03.2007
// Voice memory option installed flag added 17.03.2007

#include <stdio.h>					// Added - 5B4AIY
#include <ctype.h>
#include "juma-trx2.h"				// Board definitions
#include "trx2_eeprom.h"			// Get EEPROM structure definitions

// User Configuration Data
extern struct
	{
	struct defval defval;
	} eeprom;

// System Calibration Data
extern struct
	{
	struct calval calval;
	} cal;

// External References
// External Functions
// UART1 Functions
extern int kbhit(void);
extern unsigned char getch(void);

// UART2 Functions
extern int kbhit2(void);
extern unsigned char getch2(void);

// LCD Functions
extern void display_screen(const char *, const char *);
extern void lcd_putst(register const char *);
extern void lcd_cmd(unsigned char);

// General Timing Function
extern void ms_delay(int);

// ISD1700 Test
extern unsigned int tx_rx_16_spi2(unsigned int);
extern void set_isd_addresses(int);
extern void play_mem(int, int);
extern void rec_mem(int, int);

// Sound Effect
extern void beep(int, int);						// Generate beep, period, length

// Misc External Data
extern int isd_1700_installed;					// Voice memory board found
extern int tx_enable;							// 1 = Enable, 0 = Disable
extern int vfo_idx;
extern const char out_of_band[];				// "Out Of Band!" error message
extern const char PWR_No_MODE_Yes[];
extern const char *saved_restored[];
extern const char *save_settings_prompt[];

// Module Variables
static int voice_mem;							// Selected voice memory bank, 0-9
static unsigned char cmd;						// Stored command
const char mode_error[] = {" VM Mode Error! "};
const char reserved_memory[] = {"Reserved Memory!"};

void Display_Msg(const char *msg, int line)
	{
	lcd_cmd((line) ? LINE2 : LINE1);
	lcd_putst(msg);
	beep((line) ? HZ587_31 : HZ698_45, cal.calval.beep_len);
	ms_delay(1200);
	}

// Voice Memory Command Parser - called by serial_ext_kb() and uart2_ext_kb()
void _VM_Command(int _cmd)
	{
	const char *msg;

	switch(_cmd)
		{
// Store valid commands
		case 'M':									// Record from microphone
		case 'P':									// Play selected memory
		case 'T':									// Transmit from selected memory
			beep(HZ587_31, cal.calval.beep_len);	// Acknowledge command
			cmd = _cmd;								// Save selected command
		break;

// Service commands, execute immediately
		case 'R':									// Immediate record from receiver audio
			beep(HZ587_31, cal.calval.beep_len);	// Acknowledge command
			rec_mem(0, 1);							// Record from receiver audio output
		break;

		case 'S':
		default:
			beep(HZ587_31, cal.calval.beep_len);	// Acknowledge command
			tx_rx_16_spi2(0x0002 | 0x0010);
		break;

		case 'E':
			display_screen(save_settings_prompt[2], PWR_No_MODE_Yes);
			beep(HZ587_31, cal.calval.beep_len);	// Acknowledge command

			while(SW5 && !PWR_SW);					// Wait for a MODE or PWR button push

			if(_MODE)								// If MODE button was pressed,
				{
				tx_rx_16_spi2(0x0043 | 0x0010);
				msg = saved_restored[2];			// 'Erased!' message
				}
			else
				msg = saved_restored[0];			// 'Cancelled!' message

			Display_Msg(msg, 1);					// Display message on line 2
		break;

// Select memory bank
// 0 = Long memory reserved for recording from RX, others are for short messages
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			voice_mem = _cmd - '0';
// Start stored command
			switch(cmd)
				{
// ISD1700 commands
// Record from microphone
				case 'M':
					if(voice_mem)
						rec_mem(voice_mem, 0);		// Only permit microphone recordings into M1 - M9
					else
						Display_Msg(reserved_memory, 0);	// Display 'Reserved Memory' error message on line 1
				break;
// Play to speaker
				case 'P':
					play_mem(voice_mem, 0);
				break;
// TX on the air
				case 'T':
					if(eeprom.defval.mode[vfo_idx] > USB)
						{
						Display_Msg(mode_error, 0);	// Display 'VM Mode Error' error message on line 1
						break;
						}

					if(tx_enable)
						{
						ISD_PLAY_BAND = 1;			// Connect ISD1700 to MIC channel
						PTT_OUT = 1;				// TX on
						play_mem(voice_mem, 1);		// Play message, stop at EOM or end address
						PTT_OUT = 0;				// TX off
						ISD_PLAY_BAND = 0;			// Disconnect ISD1700 from MIC channel
						}
					else
						Display_Msg(out_of_band, 0);// Display 'Out Of Band!' error message on line 1
				break;
				}	// End switch(cmd)
		break;
		}		// End switch(_cmd)
	}

// UART1 Serial Port External Keyboard/Terminal Mode Handler Function
void serial_ext_kb(void)
	{
	if(kbhit()) _VM_Command(toupper(getch()));
	}

// To avoid a possible Voice Memory command conflict, this function will only execute if UART1
// is not selected as the Voice Memory control. If it is, then any characters received on UART2
// will be discarded as they will if the VM is not installed.
// UART2 Serial Port External Keyboard/Terminal Mode Handler Function
void uart2_ext_kb(void)
	{
	char c;

	if(kbhit2())
		{
		c = toupper(getch2());

		if((Serial_Mode != SERIAL_VOICE_MEMORY) && isd_1700_installed)
			_VM_Command(c);
		}
	}

