// JUMA-TRX2 SPI 1 driver for 74HC595 shift registers
// Two 74HC595 can be connected in cascade
// Juha Niinikoski, OH2NLT 03.12.2006
 
// MPLAB C30 version

// MCU dsPIC30F6014A

// Quiet bus test to reduce main board interference 23.12.2006
// Introduced an audio mute into function exchg_data_spi1() to eliminate click when changing modes. - 5B4AIY
// Changed old_data to be a static unsigned int in exchg_data_spi1() rather than static local. - 5B4AIY
// Removed local variable spi_rx_data_available never used. - 5B4AIY
/*
 SPI Bus Operation
 The Serial Peripheral Interchange bus is an intrinsic part of the dsPIC30F6014 microprocessor. The chip is equipped
 with two such busses, only SPI1 is used in the Juma TRX-2. For full details of the theory of operation of this
 interface bus, please refer to the Microchip dsPIC30F6014 data sheet. The following is an abbreviated description.
 The SPI bus is controlled by means of configuration settings strobed into the SPI1CON status register. In this
 application the bus is configured as a Master, meaning that it generates the necessary clock and latch strobe signals
 for the peripheral device. The control register is set to deliver 16 bits of serial data from the SPI1SR to the
 SPI1DO pin. This occurs whenever data is sent to the SPI1BUF register. At the conclusion of this operation, the SPI
 interrupt is generated, and the interrupt code is required to deliver a latch pulse to the peripheral, as well as
 read the SPI1BUF input register. In this application, no actual data is input to the microprocessor, but to ensure
 that further data can be transmitted, this dummy read operation has to be performed in order to clear the SPI1POV
 flag. The SPI1DO pin is connected to the data input (Pin 14) of IC6 on the main board, a 74HC595 serial to parallel
 converter chip. The serial data output (Pin 9 - Cascade) is connected to pin 14 of IC4 on the RF Filter board. The 16
 data bits are clocked by the inverted output of the SPI1_CLK which is itself configured to deliver negative going
 clock pulses via the SPI1CON register during the initialisation phase. (See comments of init_spi1()) At the conclusion
 the first 8 bits will have been clocked through to IC4 to set the band select filter, the second 8 bits will have been
 clocked into IC6, the control register to select filter width, AGC speed, MIC/LINE, USB/LSB/CW/Tune etc. Also at the
 conclusion the SPI1 interrupt will occur, which generates the latch pulse to latch the clocked data into the output
 registers of IC6 and IC4, as well as clear the SPI1OV flag by reading SPI1BUF in preparation for the next data output,
 and finally resetting the IRQ flag. The data is only output if there has been a change, this minimises the audio
 interference that would result from a continuous clocking of the SPI bus. A.Ryan - 5B4AIY - 09/JAN/2014
*/

#include "juma-trx2.h"				// JUMA-TRX2 hardware specific definitions
#define SPI1_LATCH _LATB2			// Strobe signal in I/O pins

// External functions
extern void us_delay(unsigned int);	// Delay routines
extern void ms_delay(unsigned int);

// Init SPI 1
// SPI1CON = 0x0439
// 04 = Set Bit 10, 16 bit operation
// 39 = Set Bit 5, Master Enable, Bit 4 SPRE2, Bit 3, SPRE1, Bit 0, PPRE0
// SPI1STAT = 0x8000 Set Bit 15, SPI1 Enable
void init_spi1(void)
	{
	SPI1_LATCH = 0;					// Set Slave Latch to idle state, low
	SPI1CON = 0x0439;				// Bit clock falling edge, clock / 32
	SPI1STAT = 0x8000;				// SPI1 ON
	_SPI1IF = 0;					// Clear IRQ flag
	_SPI1IE = 1;					// Enable SPI TX ready IRQ
	}
/*
 Send Data
 This routine was modified by the commented code to eliminate a loud click every time the mode was
 changed when switching from LSB > USB, from USB > CW. The cure is to mute the audio whilst the data
 is being changed. This produces a brief silence which is far preferable to a loud click. - 5B4AIY
 This mute is now executed in the calling function. A.Ryan - 5B4AIY
*/
void exchg_data_spi1(unsigned int new_data)
	{
   	while(SPI1STATbits.SPITBF);		// Wait until 'SPITBF' bit is cleared

	SPI1BUF = new_data;
	}
/*
 TX ready, generate latch pulse, clear the SPI1OV flag, clear the IRQ flag.
 After a considerable amount of testing, incrementing a variable if the flag was set, and another if the flag was clear,
 it was determined that the SPI1STATbits.SPIRBF flag was always set at the end of a data transfer, and thus the read
 operation to clear the SPI1OV flag is always required. Consequently, the test is redundant, as it will always be TRUE.
 However, if it is removed, then 5 assembly language NOP instructions would be required to insert the equivalent delay
 to ensure that the latch pulse width remains the same. The test is therefore retained merely to maintain the width of
 the latch pulse at about 1.2uSec, even though subsequent testing shows that the transceiver works perfectly well even
 if the latch pulse is shortened by 533nS, the execution time for 5 instructions. A.Ryan - 5B4AIY - 12/JAN/2014

 The width of the latch pulse was measured at 1.2uSec at the half-amplitude points, where a gate would trigger using a
 Tektronix TDS-1012 Digital Storage Oscilloscope. To minimise capacitive loading, a x10 probe was used. When the redundant
 test was commented out, the pulse width was 500nSec, with a peak amplitude of 4V. Although this worked perfectly well,
 it was obviously right on the tolerance limit. The previous pulse width was obtained when 5 assembly language NOP
 instructions were inserted. Since there was no real performance gain, the original redundant code was restored.
 There is obviously a fair amount of capacitive loading on the drive signal, as the rise and fall-times are of the order
 of 400nSec. Nevertheless, the pulse does reach +5V, and the width is more than sufficient. A.Ryan - 5B4AIY - 06/SEP/2014
*/
void __attribute__((interrupt)) _SPI1Interrupt(void)
	{
	unsigned int spi_rx_data;

	SPI1_LATCH = 1;					// Generate latch strobe pulse __-__, rising edge sensitive

	if(SPI1STATbits.SPIRBF)			// Redundant test, always TRUE - retained to preserve latch pulse width.
		spi_rx_data = SPI1BUF;		// Dummy read RX data, to clear internal SPI1OV flag

	SPI1_LATCH = 0;					// Back to idle state
	_SPI1IF = 0;					// Clear IRQ flag
	}

