// Simple UART2 interface for JUMA-TRX2
// Juha Niinikoski, OH2NLT 01.02.2008

// Data type cast corrections 18.11.2008
// Rewrote putst2() in a more compact form - 5B4AIY 29/AUG/11

#include "juma-trx2.h"
#include <stdio.h>

// Module variables & definitions
static volatile unsigned char rx_buffer2[256];		// UART mod(256) RX queue, must be exactly 256 long
static volatile unsigned char rx_buf_out_idx2 = 0;	// Buffer output index, pointer behind in idx
static volatile unsigned char rx_buf_in_idx2 = 1;	// Buffer input index, pointer always "leading"

// Baud Rate divisors for 30 MHz osc =  7.5 MHz FCY 
// Divisor = ((FCY / (Baudrate * 16)) - 1
const unsigned int u2_baud_rates[] = {
										390,	// 0 - 1200	Baud	389.625
										194,	// 1 - 2400	Baud	194.3125
										97,		// 2 - 4800	Baud	96.65625
										48,		// 3 - 9600	Baud	47.828125
										23,		// 4 - 19200 Baud	23.4140625
										11,		// 5 - 38400 Baud	11.20703125
										7,		// 6 - 57600 Baud	7.138020833
										3		// 7 - 115200 Baud	3.069010417
									};

// UART2 IRQ service
// Queue character
void __attribute__((interrupt)) _U2RXInterrupt(void)
	{
  	unsigned char rx;

  	rx = U2RXREG;							// Data read

  	if((unsigned char)(rx_buf_in_idx2 + 1) != rx_buf_out_idx2)	// Check overrun
  		{
 		rx_buffer2[rx_buf_in_idx2] = rx;	// Place just received character in queue
		rx_buf_in_idx2++;
    	}
// Else reject character
// Check & handle possible errors
	if(U2STAbits.OERR == 1)					// OERR is blocking UART, should not happen but...
		U2STAbits.OERR = 0;					// Reset overrun error if occured

// This is not automatic in 30F6014 UART
	IFS1bits.U2RXIF = 0;					// Reset IRQ flag
	}

// Clear serial port, reset input buffer
void ClearUSART2queue(void)
	{
	rx_buf_out_idx2 = 0;					// Buffer output index
	rx_buf_in_idx2 = 1;						// Buffer input index
	}

// UART2 init
void InitUSART2(void)
	{
	U2BRG = u2_baud_rates[3];				// Set Baud rate to 9600
	U2MODE = 0x0000;
	U2STA = 0x0400;							// IRQ on every character, enable transmitter
 	IFS1bits.U2RXIF = 0;					// Clear possible pending IRQ
 	IEC1bits.U2RXIE = 1;					// Enable RX IRQs
 	U2MODEbits.UARTEN = 1;					// Enable UART
	U2STAbits.UTXEN = 1;					// Enable transmitter, UART2 needs this after UART enable !
	ClearUSART2queue();						// Ensure buffer pointers are reset.
	}

// Set UART2 baud rate
void SetUSART2baud(int baud_sel)
	{
	if(baud_sel > 7) baud_sel = 7;			// Stay in the table

	 U2BRG = u2_baud_rates[baud_sel];		// Set Baud rate
	}

// UART2 print character
void putch2(unsigned char c)
	{
	while(U2STAbits.UTXBF == 1);			// Wait here if TX FIFO full

	U2TXREG = (unsigned int)c;				// Tx reg is 16-bits long
	}

// UART2 print null terminated string
void putst2(const char *str)
	{
	while(*str) putch2(*str++);				// More concise form. - 5B4AIY
	}

// Check if key pressed, data available from receive buffer
// Not pressed = 0, Pressed = number of characters
int kbhit2(void)
	{
	return(rx_buf_in_idx2 - (unsigned char)(rx_buf_out_idx2 + 1));	// Return with number of available characters
	}

// Get character from RX queue, IRQ RX system
unsigned char getch2(void)
	{										// The compiler produces exactly the same code for: while(kbhit2() == 0);
 	while(!kbhit2());						// Wait for data if queue is empty 
  	return rx_buffer2[++rx_buf_out_idx2];	// Return with data
	}

// Receive character with echo
unsigned char getche2(void)
	{
	unsigned char c;

	putch2(c = getch2());
	return c;
	}

